import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart';

typedef OnInitCallback<M> = void Function(M manager);
typedef InitialDataCreator<M, T> = T Function(M manager);
typedef StreamCreator<M, T> = Stream<T> Function(M manager);
typedef Builder<M, T> = Widget Function(
  BuildContext,
  M manager,
  T model,
);
typedef OnDisposeCallback<M> = void Function(
  M manager,
);
typedef OnWillChangeCallback<M> = void Function(M manager);
typedef OnInitialBuildCallback<M> = void Function(M manager);

class StatefulStreamBuilder<M, T> extends StatelessWidget {
  final Builder<M, T> builder;
  final InitialDataCreator<M, T> initialData;
  final StreamCreator<M, T> stream;
  final OnInitCallback<M> onInit;
  final OnDisposeCallback<M> onDispose;
  final OnWillChangeCallback<M> onWillChange;
  final OnInitialBuildCallback<M> onInitialBuild;
  StatefulStreamBuilder({
    Key key,
    @required this.builder,
    @required this.stream,
    this.initialData,
    this.onInit,
    this.onDispose,
    this.onWillChange,
    this.onInitialBuild,
  })  : assert(builder != null),
        super(key: key);

  @override
  Widget build(BuildContext context) {
    return _StatefulLifecycleManager<M, T>(
      builder: builder,
      manager: Provider.of<M>(context, listen: false),
      stream: stream,
      initialData: initialData,
      onInit: onInit,
      onDispose: onDispose,
      onWillChange: onWillChange,
      onInitialBuild: onInitialBuild,
    );
  }
}

class _StatefulLifecycleManager<M, T> extends StatefulWidget {
  final M manager;
  final Builder<M, T> builder;
  final InitialDataCreator<M, T> initialData;
  final StreamCreator<M, T> stream;
  final OnInitCallback<M> onInit;
  final OnDisposeCallback<M> onDispose;
  final OnWillChangeCallback<M> onWillChange;
  final OnInitialBuildCallback<M> onInitialBuild;

  _StatefulLifecycleManager({
    Key key,
    @required this.builder,
    @required this.stream,
    @required this.manager,
    this.initialData,
    this.onInit,
    this.onDispose,
    this.onWillChange,
    this.onInitialBuild,
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() {
    return _StatefulLifecycleManagerState<M, T>();
  }
}

class _StatefulLifecycleManagerState<M, T>
    extends State<_StatefulLifecycleManager<M, T>> {
  @override
  void initState() {
    super.initState();
    if (widget.onInitialBuild != null) {
      WidgetsBinding.instance.addPostFrameCallback((_) {
        widget.onInitialBuild(widget.manager);
      });
    }

    if (widget.onInit != null) {
      widget.onInit(widget.manager);
    }
  }

  @override
  void dispose() {
    super.dispose();

    if (widget.onDispose != null) {
      widget.onDispose(widget.manager);
    }
  }

  @override
  Widget build(BuildContext context) {
    return StreamBuilder<T>(
      initialData: widget.initialData != null
          ? widget.initialData(widget.manager)
          : null,
      stream: widget.stream(widget.manager),
      builder: (ctx, snapshot) =>
          widget.builder(ctx, widget.manager, snapshot.data),
    );
  }
}
